//
//  IMplementation of the DSfilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/


#include <sstream>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <DataSubspace.h>
#include <errstr.h>


const DSS::DSfilterComp DSS::noComps(0);

const std::string DSfilter::comment("data subspace descriptor: ");


DSfilter::DSfilter(DScomp *parent,
		   const FilterType ftype, const std::string &name, 
		   const std::string &units /*= ""*/,
		   const DSfilterComp &filtercomp /*= noComps*/) :
  DSchild<DScomp>(parent), 
  _ftype(ftype), _name(name), _units(units), _comps(filtercomp), _identifier(0) {}


DSfilter::DSfilter(DScomp *parent,
		   const FilterType ftype, const Column *inheritedColumn) : 
  DSchild<DScomp>(parent), _ftype(ftype), _identifier(0) {

  _name = inheritedColumn->name();
  _units = inheritedColumn->units();
}


DSfilter::FilterType DSfilter::type() const{
  return _ftype;
}


std::string DSfilter::name() const{
  return _name;
}


std::string DSfilter::units() const{
  return _units;
}


DSS::DSfilterComp DSfilter::comps() const{
  DSS::DSfilterComp lcomps;

  if( _comps!=noComps ){
    lcomps = _comps;
  }else{
    if( _name=="POS" ){
      lcomps.push_back("X");
      lcomps.push_back("Y");

    }else if( _name=="EQPOS" ){
      lcomps.push_back("RA");
      lcomps.push_back("DEC");

    }else if( _name=="GALPOS" ){
      lcomps.push_back("GLON");
      lcomps.push_back("GLAT");

    }else{
      lcomps.push_back(_name+"X");
      lcomps.push_back(_name+"Y");
    }
  }

  return lcomps;
}


bool DSfilter::like(const DSfilter &ofilter) const{
  if( ofilter._name==_name     &&
      ofilter._ftype==_ftype   &&
      ofilter._units==_units   &&
      ofilter.comps()==comps() && 
      ofilter._identifier==_identifier)
    return true;

  return false;
}



std::list<Range<double> > DSfilter::ranges() const{
  errstr << "No range information available for this filter type." 
	 << Sas::error("NoRange");

  std::list<Range<double> > dummy;
  return dummy;
}


Range<double> DSfilter::extent(const std::string &axis) {
  if( axis!=_name )
    errstr << "Attempt to extract extent information from an unsupported filter"
	   << Sas::error("DSextentType");

  Range<double> r(0,0,Unbounded,Unbounded);
  return r;
}

 Range<double> DSfilter::realextent(const std::string &axis) {
  if( axis!=_name )
    errstr << "Attempt to extract extent information from an unsupported filter"
	   << Sas::error("DSextentType");

  Range<double> r(0,0,Unbounded,Unbounded);
  return r;
}

vector<unsigned long> DSfilter::onBitMasks() const{
  errstr << "No bit mask information available for this filter type." 
	 << Sas::error("NoMask");

  vector<unsigned long> dummy;
  return dummy;
}


vector<unsigned long> DSfilter::offBitMasks() const{
  errstr << "No bit mask information available for this filter type." 
	 << Sas::error("NoMask");

  vector<unsigned long> dummy;
  return dummy;
}


GeoRegion DSfilter::region() const{
  errstr << "No region information available for this filter type." 
	 << Sas::error("NoRegion");

  GeoRegion dummy;
  return dummy;
}


std::string DSfilter::blockSpec() const{
  errstr << "No block information available for this filter type." 
	 << Sas::error("NoBlock");

  return " ";
}

void DSfilter::operator!() {
  errstr << "Unable to perform negation of this filter type."
	 << Sas::error("NoNegation");
}

void DSfilter::_getRefBlockName(unsigned compNumber, unsigned filterNumber, 
				Block *toBlock, std::string blockSpec, std::string &ref) const{
  std::string refValue;

  while(true){
    refValue = ":" + ref;

    // Need to copy the filter block to the output data set.
    // First check for a name clash

    bool prevRef = false;
    for(unsigned iComp=1; iComp<compNumber; iComp++){
      std::stringstream pRefName;
      if( iComp==1 )
	pRefName << "DSREF" << filterNumber;
      else
	pRefName << iComp << "DSREF" << filterNumber;

      if( toBlock->hasAttribute(pRefName.str()) &&
	  toBlock->attribute(pRefName.str())->asString() == refValue ){

	if( parent()->parent()->component(iComp-1)
	    ->filter(*this)->blockSpec() == blockSpec ) prevRef = true;
      }
    }

    if( toBlock->parent()->hasBlock(ref) ){
      if( toBlock->parent()->name()+refValue != blockSpec && !prevRef ){
	errstr << "Name clash while writing Data Subspace: Block "
	       << ref << " already exists in output data set. ";

	std::string nameRoot;
	if( type()==GtiT ) nameRoot="GTI";
	else if( type()==RegionBlockT || type()==RegionT ) nameRoot="REG";
	else nameRoot="BLK";

	std::stringstream blkNumStr;
	blkNumStr << nameRoot;
	blkNumStr.width(5);
	blkNumStr.fill('0');
	if( ref.substr(0,3) == nameRoot ){
	  int blkNum = atoi( ref.substr(3,5).c_str() ) + 1;
	  blkNumStr << blkNum;
	}else{
	  blkNumStr << compNumber*100+filterNumber;
	}
	ref = blkNumStr.str();

	errstr << "Renaming block to " << ref << "." << Sas::warning("DssBlockClash");
      }else{
	return;
      }
    }else{
      return;
    }
  }
  
}


GeoRegion DSfilter::_negateRegion(GeoRegion &region){
  unsigned nComps = region.size();
  if(nComps==0) return region;

  GeoRegion newRegion;
  std::vector<unsigned> shapeCount, compSize;
  
  for(unsigned icomp=0; icomp<nComps; icomp++){
    shapeCount.push_back(0);
    compSize.push_back(region[icomp].size());
  }

  while(true){
    Component comp;
    for(unsigned icomp=0; icomp<nComps; icomp++){
      FShape *newShape, *shape=region[icomp][shapeCount[icomp]];
      switch( shape->type() ){
      case ShapeBasics::POINT:
	newShape = new FShape(*shape);
	break;
      case ShapeBasics::LINE:
	newShape = new FLine(dynamic_cast<FLine &>(*shape));
	break;
      case ShapeBasics::CIRCLE:
	newShape = new FCircle(dynamic_cast<FCircle &>(*shape));
	break;
      case ShapeBasics::SECTOR:
	newShape = new FSector(dynamic_cast<FSector &>(*shape));
	break;
      case ShapeBasics::ANNULUS:
	newShape = new FAnnulus(dynamic_cast<FAnnulus &>(*shape));
	break;
      case ShapeBasics::ELLIPSE:
	newShape = new FEllipse(dynamic_cast<FEllipse &>(*shape));
	break;
      case ShapeBasics::ELLIPTANNULUS:
	newShape = new FElliptAnnulus(dynamic_cast<FElliptAnnulus &>(*shape));
	break;
      case ShapeBasics::BOX:
	newShape = new FBox(dynamic_cast<FBox &>(*shape));
	break;
      case ShapeBasics::ROTBOX:
	newShape = new FRotBox(dynamic_cast<FRotBox &>(*shape));
	break;
      case ShapeBasics::RECTANGLE:
	newShape = new FRectangle(dynamic_cast<FRectangle &>(*shape));
	break;
      case ShapeBasics::ROTRECTANGLE:
	newShape = new FRotRectangle(dynamic_cast<FRotRectangle &>(*shape));
	break;
      case ShapeBasics::RHOMBUS:
	newShape = new FRhombus(dynamic_cast<FRhombus &>(*shape));
	break;
      case ShapeBasics::ROTRHOMBUS:
	newShape = new FRotRhombus(dynamic_cast<FRotRhombus &>(*shape));
	break;
      case ShapeBasics::POLYGON:
	newShape = new FPolygon(dynamic_cast<FPolygon &>(*shape));
	break;
      }
      comp.push_back(newShape);
    }
    newRegion.push_back(comp);
    
    shapeCount[0]++;
    for(unsigned icomp=0; icomp<nComps-1; icomp++){
      if(shapeCount[icomp]==compSize[icomp]){
	shapeCount[icomp]=0;
	shapeCount[icomp+1]++;
      }else break;
    }
    if(shapeCount[nComps-1]==compSize[nComps-1]) break;
  }

  for(unsigned icomp=0; icomp<newRegion.size(); icomp++){
    for(unsigned ishape=0; ishape<newRegion[icomp].size(); ishape++)
      newRegion[icomp][ishape]->include( !newRegion[icomp][ishape]->included() );
  }

  return newRegion;
}


