//
//  IMplementation of the Range class and associated utilities
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <cmath>
#include <cfloat>
#include <cstring>

#include <DataSubspace.h>


double num_limits<double>::max = DBL_MAX;
long num_limits<long>::max = LONG_MAX;

template<class T> Range<T>::Range(bool nullify) 
  : _lower(0), _upper(0), _lowerB(Unbounded), _upperB(Unbounded), _null(nullify) {
  if(nullify){
    _lower = _upper = num_limits<T>::max;
    _lowerB = _upperB = Closed;
  }
}

template<class T> template<class U> Range<T>::Range(const Range<U> &inRange) 
  : _null(false) {
  U lower, upper;
  RangeBound lowerB, upperB;
  inRange.values(lower, upper, lowerB, upperB);

  if(lower == -num_limits<U>::max) _lower = -num_limits<T>::max;
  else _lower = static_cast<T>(lower);
  if(upper == num_limits<U>::max) _upper = num_limits<T>::max;
  else _upper = static_cast<T>(upper);
  _lowerB = lowerB;
  _upperB = upperB;  
}


template<class T> Range<T>::Range(T lower, T upper, 
				  RangeBound lowerB, RangeBound upperB) :
  _lower(lower), _upper(upper), _lowerB(lowerB), _upperB(upperB), _null(false) {}


template<class T> template<class U> Range<T>::Range(U lower, U upper,
				  RangeBound lowerB, RangeBound upperB) :
  _lowerB(lowerB), _upperB(upperB), _null(false) {
    
    _lower = lower;
    _upper = upper;
}


template<class T> void Range<T>::values(T &lower, T &upper,
			       RangeBound &lowerB, RangeBound &upperB) const{
  lower = _lower;
  upper = _upper;
  lowerB = _lowerB;
  upperB = _upperB;
}


template<class T> bool Range<T>::contains(T value) const{
  if( _lowerB!=Unbounded ){
    if( value < _lower ) return false;
    if( _lowerB == Open && value == _lower ) return false;
  } 
  if( _upperB!=Unbounded ){
    if( value > _upper ) return false;
    if( _upperB == Open && value == _upper ) return false;
  } 

  return true;
}


template<class T> template<class U> bool 
Range<T>::operator<(const Range<U>& inRange) const{
  U lower, upper;
  RangeBound lowerB, upperB;
  
  inRange.values(lower, upper, lowerB, upperB);

  // First sort based on the lower limit.

  if( _lowerB!=Unbounded && lowerB!=Unbounded ){

    if( _lower < lower ) return true;
    if( _lower > lower)  return false;

    // A closed interval is taken to be less than an equivalent open interval.

    if( _lowerB==Closed && lowerB==Open ) return true;
    if( _lowerB==Open && lowerB==Closed ) return false;

  }else{

    // an undefined lower limit is less than a specified one.

    if( _lowerB==Unbounded && lowerB!=Unbounded ) return true;
    if( _lowerB!=Unbounded && lowerB==Unbounded ) return false;
  }

  // Lower limits are equivalent, so now sort on upper limits.

  if( _upperB!=Unbounded && upperB!=Unbounded ){

    if( _upper < upper ) return true;
    if( _upper > upper)  return false;

    // A closed interval is taken to be greater than an equivalent 
    // open interval.

    if( _upperB==Closed && upperB==Open ) return false;
    if( _upperB==Open && upperB==Closed ) return true;

  }else{

    // an undefined upper limit is greater than a specified one.

    if( _upperB==Unbounded && upperB!=Unbounded ) return false;
    if( _upperB!=Unbounded && upperB==Unbounded ) return true;
  }

  // The two ranges are identical

  return false;
}


template<class T> template<class U> bool 
Range<T>::operator==(const Range<U>& inRange) const{
  U lower, upper;
  RangeBound lowerB, upperB;
  bool lowMatch=true, highMatch=true;
  
  inRange.values(lower, upper, lowerB, upperB);

  if( lowerB!=_lowerB || upperB!=_upperB ) return false;

  if( lowerB!=Unbounded ) lowMatch = lower==_lower;

  if( upperB!=Unbounded ) highMatch = upper==_upper;

  return highMatch && lowMatch;
}


template<class T> template<class U> bool 
Range<T>::operator!=(const Range<U>& inRange) const{
  return !(*this==inRange);
}


template<class T>
std::list<Range<T> > parseRanges(const std::string &inString){
  std::list<Range<T> > outRanges;
  std::string parseString = inString;

  RangeBound upperB, lowerB;
  T upper = 0, lower = 0;

  while( parseString.size() != 0 ){
    std::string::size_type icomma = parseString.find(",");
    if( icomma == std::string::npos ) icomma = parseString.size();
    std::string rangeString = parseString.substr(0,icomma);
    parseString.erase(0,icomma+1);

    switch( *rangeString.substr(0,1).data() ){
    case '[':
      lowerB = Closed;
      rangeString.erase(0,1);
      break;
    case '(':
      lowerB = Open;
      rangeString.erase(0,1);
      break;
    case ':':
      lowerB = Unbounded;
      break;
    default :
      lowerB = Closed;
    }

    std::string::size_type iparse = rangeString.size();
    switch( *rangeString.substr(iparse-1,1).data() ){
    case ']':
      upperB = Closed;
      rangeString.erase(iparse-1,1);
      break;
    case ')':
      upperB = Open;
      rangeString.erase(iparse-1,1);
      break;
    case ':':
      upperB = Unbounded;
      break;
    default :
      upperB = Closed;
    }

    std::string::size_type icolon = rangeString.find(":");
    if( icolon==std::string::npos ){
      upper = lower = (T) atof(rangeString.c_str());
    }else{
      if( icolon!=0 )
	lower = (T) atof(rangeString.substr(0,icolon).c_str());
      if( icolon!=rangeString.size()-1 )
	upper = (T) atof(rangeString.substr(icolon+1).c_str());
    }

    Range<T> tempRange(lower, upper, lowerB, upperB);
    outRanges.push_back(tempRange);
  }

  return outRanges;
}

template<class T>
Range<T> maxRange(const std::list<Range<T> > &ranges){
  T upper, lower, upperMax, lowerMax;
  RangeBound upperB, lowerB, upperBmax, lowerBmax;

  typename std::list<Range<T> >::const_iterator irange;

  if( ranges.size()==0 ){
    Range<T> max(0, 0, Unbounded, Unbounded);
    return max;
  }

  ranges.begin()->values(lowerMax, upperMax, lowerBmax, upperBmax);

  for(irange=ranges.begin()++; irange!=ranges.end(); irange++){
    irange->values(lower, upper, lowerB, upperB);

    if( lowerB == Unbounded ) lowerBmax = Unbounded;
    else if( lower == lowerMax ){
      if( lowerB == Closed ) lowerBmax = Closed;
    }else if( lower < lowerMax ){
      lowerMax = lower;
      lowerBmax = lowerB;
    }

    if( upperB == Unbounded ) upperBmax = Unbounded;
    else if( upper == upperMax ){
      if( upperB == Closed ) upperBmax = Closed;
    }else if( upper > upperMax ){
      upperMax = upper;
      upperBmax = upperB;
    }
  }

  Range<T> max(lowerMax, upperMax, lowerBmax, upperBmax);
  return max;
}

template<class T>
std::list<Range<T> > operator&(std::list<Range<T> > rl1, std::list<Range<T> > rl2){
  typename std::list<Range<T> >::const_iterator ir1, ir2;
  std::list<Range<T> > newrl;
  T low1=0, low2=0, up1=0, up2=0, newlow=0, newup=0;
  RangeBound lb1=Closed, lb2=Closed, ub1=Closed, ub2=Closed, nub=Closed, nlb=Closed;

  rl1.sort();
  rl2.sort();
  
  ir2 = rl2.begin();
  ir2->values(low2, up2, lb2, ub2);

  for( ir1=rl1.begin(); ir1!=rl1.end(); ir1++){
    bool done=false;
    ir1->values(low1, up1, lb1, ub1);
    // Find ranges where the upper value for range2 >= lower value for range1
    while(ir2!=rl2.end()){
      if(ub2!=Unbounded && lb1!=Unbounded){
	if(up2>low1) break;
	if(up2==low1 && (ub2==Closed || ub2==lb1) ) break;
      }else break;
      ir2++;
      ir2->values(low2, up2, lb2, ub2);
    }
    // Is lower value of range 2 > the upper value of range1?
    while(ir2!=rl2.end()){

      if(lb2!=Unbounded && ub1!=Unbounded){
	if(low2>up1) break;
	if(low2==up1 && (lb2==Open || lb2!=ub1) ) break;
      }

      // Overlap found      
      if(low1>low2){
	newlow=low1;
	nlb=lb1;
      }else if(low1<low2){
	newlow=low2;
	nlb=lb2;
      }else{
	newlow=low1;
	if(lb1==Open || lb2==Open) nlb=Open;
	else if(lb1==Closed || lb2==Closed) nlb=Closed;
	else nlb=Unbounded;
      }

      if(up1<up2 || ub2==Unbounded){
	newup=up1;
	nub=ub1;
	done=true;
      }else if(up1>up2 || ub1==Unbounded){
	newup=up2;
	nub=ub2;
	ir2++;
      }else{
	newup=up1;
	if(ub1==Open){
	  nub=Open;
	  done=true;
	}else if(ub2==Open){
	  nub=Open;
	  ir2++;
	}else if(ub1==Closed){
	  nub=Closed;
	  done=true;
        }else if(ub2==Closed){
	  nub=Closed;
	  ir2++;
	}else{
	  nub=Unbounded;
	  done=true;
	}
      }

      ir2->values(low2, up2, lb2, ub2);
      Range<T> newrange(newlow, newup, nlb, nub);
      newrl.push_back(newrange);
      if(done){
	ir2 = rl2.begin();
	break;
      }
    }
  }

  if(newrl.size()==0){
    Range<T> newrange(true);
    newrl.push_back(newrange);    
  }
  return newrl;
}


template Range<long>::Range(bool);
template Range<double>::Range(bool);

template class Range<long>;
template class Range<double>;

template Range<long>::Range(const Range<double> &);
template Range<double>::Range(const Range<long> &);

template Range<double>::Range(int, int, RangeBound, RangeBound);
template Range<double>::Range(long, long, RangeBound, RangeBound);
template Range<double>::Range(unsigned, unsigned, RangeBound, RangeBound);
template Range<double>::Range(float, float, RangeBound, RangeBound);
template Range<double>::Range(short, short, RangeBound, RangeBound);
template Range<double>::Range(unsigned char, unsigned char, RangeBound, RangeBound);

template std::list<Range<long> > parseRanges<long>(const std::string &);
template std::list<Range<double> > parseRanges<double>(const std::string &);

template std::list<Range<long> > operator&(const std::list<Range<long> >, const std::list<Range<long> >);
template std::list<Range<double> > operator&(const std::list<Range<double> >, const std::list<Range<double> >);

template Range<long> maxRange(const std::list<Range<long> > &);
template Range<double> maxRange(const std::list<Range<double> > &);

template bool Range<long>::contains(long) const;
template bool Range<double>::contains(double) const;


template bool Range<long>::operator==(const Range<long>& inRange) const;
template bool Range<long>::operator==(const Range<double>& inRange) const;
template bool Range<double>::operator==(const Range<long>& inRange) const;
template bool Range<double>::operator==(const Range<double>& inRange) const;


template bool Range<long>::operator!=(const Range<long>& inRange) const;
template bool Range<long>::operator!=(const Range<double>& inRange) const;
template bool Range<double>::operator!=(const Range<long>& inRange) const;
template bool Range<double>::operator!=(const Range<double>& inRange) const;


template bool Range<long>::operator<(const Range<long>& inRange) const;
template bool Range<long>::operator<(const Range<double>& inRange) const;
template bool Range<double>::operator<(const Range<long>& inRange) const;
template bool Range<double>::operator<(const Range<double>& inRange) const;
